// NetHackApp.CPP
//
// Copyright (c) 1997-1999 Symbian Ltd.  All rights reserved.
//

#include "NetHackApp.h"
#include "NethackApp.rsg"
#include "nhdialog.h"
#include <eikdll.h>
#include <f32file.h>
#include <bautils.h>
#include <basched.h>
#include <eikenv.h>
#include <apgtask.h>
#include <apgwgnam.h>
#include <apacln.h>
#include <coeutils.h>
#include <eikmenup.h>
#include <eikproc.h>
#include <s32file.h>

// Local functions
//
const TInt KDefaultMinHeapSize=0x10000; // !! for now
const TInt KDefaultMaxHeapSize=0x200000; // !! for now

// Find an application file. Looks in the application directory to find it.
// Updates the filename and returns ETrue if it found it.
// Returns EFalse if it didn't.
TBool CNethackAppUi::FindFile(TDes& aFileName)
{
    TFileName appName(Application()->AppFullName());
    TParse parser;
    User::LeaveIfError(parser.Set(aFileName, &appName, NULL));
    TPtrC theFile (parser.FullName());
    if (! ConeUtils::FileExists(theFile))
    {
        // File doesn't exist on current drive. Try drive C as a
        // fallback.
        User::LeaveIfError(parser.Set(_L("C:"), &aFileName, &appName));
        theFile.Set(parser.FullName());
        if (! ConeUtils::FileExists(theFile))
            return EFalse;
    }
    aFileName = parser.FullName();
    return ETrue;
}

#if defined(__WINS__)
typedef TInt (*ExeEntryPoint)(TAny*);

TInt ExeThreadStartFunction(TAny* aParam)
{
    const TPtrC commandLine = *(const TPtrC*)aParam;
    TInt space = commandLine.Locate(' ');
    if (space < 0) space = commandLine.Length();
    TPtrC libraryName = commandLine.Left(space);

    RLibrary lib;
    TInt err = lib.Load(libraryName);
    if (err == KErrNone)
    {
        ExeEntryPoint exeFunc=(ExeEntryPoint)lib.Lookup(1);
        if (!exeFunc)
            err=KErrBadLibraryEntryPoint;
        else
            err=(*exeFunc)(aParam);
    }
    User::Free(aParam);
    return(err);
}
#endif

static TThreadId StartExeL(const TDesC& aName, const TDesC& aArgs)
{
    TParse fileName;
    TThreadId threadId;
#if defined(__WINS__)
    TPtrC path=_L("Z:\\SYSTEM\\PROGRAMS\\.DXE");
    User::LeaveIfError(fileName.Set(aName,&path,NULL));
    RThread process;
#else
    TPtrC tempDes=_L("Z:\\SYSTEM\\PROGRAMS\\.EXE");
    User::LeaveIfError(fileName.Set(aName,&tempDes,NULL));
    RProcess process;
#endif
#if defined(__WINS__)
    TBuf<KMaxFileName> threadName;
    TPtrC baseName=fileName.Name();
    TInt err;
    TInt num=0;
    do
    {
        threadName.Format(_L("%S%02d"),&baseName,num++);
        err=process.Create(threadName,ExeThreadStartFunction,KDefaultStackSize,KDefaultMinHeapSize,KDefaultMaxHeapSize,NULL);
    } while(err==KErrAlreadyExists);
    User::LeaveIfError(err);
    RHeap* heap=process.Heap();

    // Copy the command line string into the heap for the new process.
    RHeap* originalHeap=User::SwitchHeap(heap);
    HBufC* commandLine = fileName.FullName().Alloc();
    commandLine->ReAlloc(commandLine->Length() + aArgs.Length() + 3);
    if (commandLine && aArgs.Length() > 0)
    {
        commandLine->Des().Append(_L(" "));
        commandLine->Des().Append(aArgs);
    }
    User::SwitchHeap(originalHeap);

    if (!commandLine)
    {
        process.Close();
        User::LeaveNoMemory();
    }

    process.SetInitialParameter(commandLine);
    threadId = process.Id();
#else
    if (aArgs.Length() > KMaxCommandLine)
        User::LeaveIfError(KErrArgument);

    //TBuf<KMaxCommandLine> buf;
    //buf.Format(_L("%S"), &aArgs);
    TInt err=process.Create(fileName.FullName(), aArgs);
    User::LeaveIfError(err);

    // get the threadId of the main thread in the exe's process
    TFullName fullName(process.Name());
    fullName.Append(_L("::Main"));
    TFindThread fT(fullName);
    fT.Next(fullName);
    RThread thread;
    thread.Open(fT);
    threadId = thread.Id();
    thread.Close();
#endif
    process.Resume();
    process.Close();
    return threadId;
}

static TBool Exec(const TDesC& aName, const TDesC& aArgs)
{
    TRAPD(err, StartExeL(aName, aArgs));
    if (err == KErrNone)
    {
        User::After(1000000);                                                            // pause 1s
        return ETrue;
    }
    return EFalse;
}


TBool CNethackAppUi::ActivateIfRunning(const TDesC& ProgramName)
{

    TApaTaskList taskList(iEikonEnv->WsSession());
    TInt numGroups = iEikonEnv->WsSession().NumWindowGroups(0);

    // look through applist to see what we can find
    for (TInt index=0; index < numGroups; index++)
    {
        TApaTask *task= new (ELeave) TApaTask(taskList.FindByPos(index));
        CleanupStack::PushL(task);
        CApaWindowGroupName* wgName=CApaWindowGroupName::NewLC(iEikonEnv->WsSession(),task->WgId());
        TUid uid = wgName->AppUid();
        TPtrC taskCaption = wgName->Caption();
        TBool found = uid.iUid == 0 &&
                      taskCaption.FindF(ProgramName)>=0; // help app with the desired caption
        if (found) 
        {
            task->BringToForeground();
            CleanupStack::PopAndDestroy(2);                                          // wgName, task
            return ETrue;                                                                  // exit procedure
        }
        CleanupStack::PopAndDestroy(2);                                              // wgName and task
    }
    return EFalse;
}

void CNethackDocument::EmbedExistingDocL(const TFullName& aFileName)
{
    if (iEmbedList->Count() > 0)
        RemoveDocL(0);

    // create a header to store the new door in
    TPictureHeader header;

    // read the dll header of the document to be embedded
    CFileStore* store;
    CStreamDictionary *dict = Process()->ReadRootStreamLC(Process()->FsSession(), store,aFileName, EFileRead|EFileShareReadersOnly);
    CleanupStack::PushL(store);
    CPersistentStore& sourceStore =*(CPersistentStore *)store;
    TApaAppIdentifier appId = CApaProcess::ReadAppIdentifierL(sourceStore,*dict);

    // Create a new glass door with an empty document in it.
    CApaDocument *doc = NULL;
    TRAPD(err, doc=Process()->AddNewDocumentL(appId.iFullName, appId.iAppUid));
    if (err==KErrNotFound)
        CEikonEnv::Static()->LeaveWithInfoMsg(R_EMBEDDED_APP_NOT_FOUND);
    else if (err != KErrNone)
        User::Leave(err);
    TApaDocCleanupItem cleanup(CEikonEnv::Static()->Process(), doc);
    CleanupStack::PushL(cleanup);
    doc->RestoreL(sourceStore,*dict);	// Restores the document.
    doc->DetachFromStoreL();

    CApaDoor* door = CApaDoor::NewLC(*doc, TSize(32,32));
    header.iPicture = door;
    // add the new entry and tidy up
    iEmbedList->AppendL(header);
    CleanupStack::Pop(3); // iPicture, doc, store
    CleanupStack::PopAndDestroy(); // stream dictionary
    delete store;
}


void CNethackDocument::RemoveDocL(TInt index)
{
    delete EmbeddedDoor(index); // destroy the door, which in turn removes and deletes the document
    iEmbedList->Delete(index); // remove the entry from the array
}

// View a document file as though it were embedded.
void CNethackDocument::ViewDocL(const TDesC& aFileName)
{
    EmbedExistingDocL(aFileName);
    // Now do embedded editing...
    EmbeddedDoor(0)->DocumentL()->EditL(this, ETrue);
}

// End of viewing the document, now we can destroy it.
void CNethackDocument::NotifyExit(MApaEmbeddedDocObserver::TExitMode /*aMode*/)
{
    // Finished editing, so remove the embedded document
    // Finished editing, so I want to remove the document.
    // Unfortunately I can't as it would involve deleting the object
    // that called me!
    //RemoveDocL(0);
}


//
// EXPORTed functions
//

EXPORT_C CApaApplication* NewApplication()
{
    return new CNethackApplication;
}

GLDEF_C TInt E32Dll(TDllReason)
{
    return KErrNone;
}


////////////////////////////////////////////////////////////////
//
// Application class, CNethackApplication
//
////////////////////////////////////////////////////////////////

TUid CNethackApplication::AppDllUid() const
{
    return KUidNetHackApp;
}

CApaDocument* CNethackApplication::CreateDocumentL()
{
    return CNethackDocument::NewL(*this);
}


////////////////////////////////////////////////////////////////
//
// Document class, CNethackDocument
//
////////////////////////////////////////////////////////////////

CNethackDocument::CNethackDocument(CEikApplication& aApp)
: CEikDocument(aApp)
{

}

void CNethackDocument::ConstructL()
{
    iEmbedList = new(ELeave) CArrayFixFlat<TPictureHeader>(1);
}

CNethackDocument* CNethackDocument::NewL(CEikApplication& aApp)
{
    CNethackDocument* self = new (ELeave) CNethackDocument(aApp);
    CleanupStack::PushL(self);
    self->ConstructL();
    CleanupStack::Pop();
    return self;
}

CNethackDocument::~CNethackDocument()
{
    // delete all the embedded doc's first...
    if (iEmbedList)
    {
        for (TInt i=iEmbedList->Count()-1 ; i>=0 ; i--)
            delete EmbeddedDoor(i); // destroys the door (which in turn removes doc from the process and deletes it)
    }
    // then...
    delete iEmbedList;
}

CEikAppUi* CNethackDocument::CreateAppUiL()
{
    return new(ELeave) CNethackAppUi;
}


////////////////////////////////////////////////////////////////
//
// App UI class, CNethackAppUi
//
////////////////////////////////////////////////////////////////

void CNethackAppUi::ConstructL()
{
    iAppView = NULL;
    BaseConstructL();
    TBuf<20> gameCaption;
    iEikonEnv->ReadResource(iGame, R_GAME_NAME);
    iEikonEnv->ReadResource(gameCaption, R_GAME_CAPTION);
#if defined(__WINS__)
    iGame[8] = 'd';
#endif
    haveGame = FindFile(iGame);
    if (haveGame && ActivateIfRunning(gameCaption))
    {
        Exit();
        return;
    }
    iCoeEnv->ReadResource(iReadMe, R_README_NAME);
    iCoeEnv->ReadResource(iManual, R_MANUAL_NAME);
    iCoeEnv->ReadResource(iGuidebook, R_GUIDEBOOK_NAME);
    iCoeEnv->ReadResource(iHints, R_HINTS_NAME);
    iCoeEnv->ReadResource(iConfig, R_CONFIG_NAME);

    haveReadMe = FindFile(iReadMe);
    haveManual = FindFile(iManual);
    haveGuidebook = FindFile(iGuidebook);
    haveHints = FindFile(iHints);
    haveConfig = FindFile(iConfig);

    iAppView=new(ELeave) CNethackAppView;
    iAppView->ConstructL(ClientRect());

    TFileName bitmapFile;
    iCoeEnv->ReadResource(bitmapFile, R_BACKGROUND_NAME);
    if (FindFile(bitmapFile))
        iAppView->SetBitmapL(bitmapFile, EMbmBackground);

    CEikFileNameLabel* filenameLabel=(CEikFileNameLabel*) iToolBar->ControlById(ENethackCmdFileName);
    filenameLabel->UpdateL();

    if (!haveGame) DimButton(EPlayGame);
    if (!haveReadMe) DimButton(EReadMe);
    if (!haveManual) DimButton(EManual);
    if (!haveGuidebook) DimButton(EGuidebook);
    //if (!haveHints) DimButton(EHints);
}

void CNethackAppUi::DimButton(TInt aCommand, TBool dimmed/*=ETrue*/)
{
    CEikButtonBase* button = STATIC_CAST(CEikButtonBase*,iToolBar->ControlById(aCommand));
    if (dimmed != button->IsDimmed())
    {
        button->SetDimmed(dimmed);
        button->DrawDeferred();
    }
}

void CNethackAppUi::DynInitMenuPaneL(TInt aMenuId,CEikMenuPane* aMenuPane)
{
    switch(aMenuId)
    {
    case R_NETHACKAPP_EDIT_MENU:
        aMenuPane->SetItemDimmed(EEditConfig, !haveConfig);
        break;

    case R_NETHACKAPP_TOOLS_MENU:
        aMenuPane->SetItemDimmed(EPlayGame, !haveGame);
        aMenuPane->SetItemDimmed(EPlayDlg, !haveGame);
        aMenuPane->SetItemDimmed(EExploreOnly, !haveGame);
        aMenuPane->SetItemDimmed(EReadMe, !haveReadMe);
        aMenuPane->SetItemDimmed(EManual, !haveManual);
        aMenuPane->SetItemDimmed(EGuidebook, !haveGuidebook);
        aMenuPane->SetItemDimmed(EHints, !haveHints);
        break;
    default :
        break;
    }
}


CNethackAppUi::~CNethackAppUi()
{
    if (iAppView)
        delete iAppView;
}

void CNethackAppUi::OpenDocL(TDesC& aFileName)
{
    EikDll::StartDocL(aFileName);
    User::After(1000000);                                                            // pause 1s
    Exit();
}

void CNethackAppUi::HandleCommandL(TInt aCommand)
{
    switch (aCommand)
    {
    case EPlayGame:
        if (Exec(iGame, _L(""))) Exit();
        break;

    case EPlayDlg:
    {
        CEikDialog *dialog = new (ELeave) CArgsDialog(this);
        if (dialog->ExecuteLD(R_NETHACKAPP_COMMANDLINE))
        {
            if (Exec(iGame, iArgs)) Exit();
        }
        break;
    }

    case EEditConfig:
    {
        CEikDialog *dialog = new (ELeave) CEditConfigDialog(this);
        if (dialog->ExecuteLD(R_NETHACKAPP_EDITCONFIG))
        {
            // Should save file here!
        }
        break;
    }

    case EExploreOnly:
        if (Exec(iGame, _L("-X"))) Exit();
        break;

    case EReadMe:
        Document()->ViewDocL(iReadMe);
        break;

    case EManual:
        Document()->ViewDocL(iManual);
        break;
    case EGuidebook:
        //OpenDocL(iGuidebook);
        Document()->ViewDocL(iGuidebook);
        break;
    case EHints:
        OpenDocL(iHints);
        break;
    case EEikCmdExit: 
        Exit();
        break;
    }
}


////////////////////////////////////////////////////////////////
//
// Application view class, CNethackAppView
//
////////////////////////////////////////////////////////////////

void CNethackAppView::ConstructL(const TRect& aRect)
{
    _LIT(KSwiss, "Swiss");
    CreateWindowL();
    SetRectL(aRect);
    TFontSpec spec(KSwiss, 213);
    iFont=iCoeEnv->CreateScreenFontL(spec);
    iText=iCoeEnv->AllocReadResourceL(R_NETHACKAPP_TEXT);

    ActivateL();
}

// Set the background bitmap.
void CNethackAppView::SetBitmapL(const TDesC&bitmapFile, TInt index)
{
    iBgndPicture = new (ELeave) CFbsBitmap();
    if (iBgndPicture->Load(bitmapFile, index, ETrue) != KErrNone)
    {
        // Failed to load the background.
        // Tidy up and do without.
        delete iBgndPicture;
        iBgndPicture = NULL;
    }
}

CNethackAppView::~CNethackAppView()
{
    iCoeEnv->ReleaseScreenFont(iFont);
    delete iText;
    delete iBgndPicture;
}

void CNethackAppView::Draw(const TRect& /*aRect*/) const
{
    CWindowGc& gc = SystemGc();
    TRect rect = Rect();	// Default size rectangle.
    TInt xDelta=0; // for x coordinates
    TInt yDelta=0; // for y coordinates

    // Draw a frame.
    //TRect drawRect=Rect();
    //drawRect.Shrink(10,10);
    //gc.DrawRect(drawRect);

    if (iBgndPicture)
    {
        // draw the background bitmap centred on the screen.
        TSize bmpSizeInPixels = iBgndPicture->SizeInPixels();
        xDelta=(rect.Width()-bmpSizeInPixels.iWidth)/2;
        yDelta=(rect.Height()-bmpSizeInPixels.iHeight)/2;
        TPoint pos=TPoint(xDelta,yDelta); // displacement vector
        pos+=rect.iTl; // bitmap top left corner position
        gc.BitBlt(pos, iBgndPicture); // CWindowGc member function
    }
#if 0
    // Add the text caption at the bottom of the window.
    TRect textRect = Rect();
    CFont *font=iFont;
    textRect.iTl = TPoint(0, font->AscentInPixels());
    textRect.SetHeight(font->AscentInPixels());

    gc.UseFont(iFont);
    gc.DrawText(*iText, textRect, font->AscentInPixels(), CGraphicsContext::ECenter, 0);
#endif
}
