// NetHackApp.H
//
// Copyright (c) 1997-1999 Symbian Ltd.  All rights reserved.
//

////////////////////////////////////////////////////////////////////
//
// This file contains the class definitions for the NetHackApp program.
//
////////////////////////////////////////////////////////////////////
#ifndef __NetHackApp_H
#define __NetHackApp_H

#include <eikappui.h>
#include <eikapp.h>
#include <eikdoc.h>

#include <eikenv.h>
#include <eikcmds.hrh>
#include <eikfnlab.h>
#include <eiktbar.h>
// Required extra for dialogs
#include <eikdialg.h>
#include <apgdoor.h>

#include <NetHackApp.rsg>
#include "NetHackApp.hrh"


#if defined(_UNICODE)
const TUid KUidNetHackApp = {0x100013cb};
#else
const TUid KUidNetHackApp = {0x100013cb};
#endif

//
// class CNethackControl
//

class CNethackDocument;

//
// class CNethackAppView
//

class CNethackAppView : public CCoeControl, public MCoeControlBrushContext
{
public:
    ~CNethackAppView();
    void ConstructL(const TRect& aRect);
    void SetBitmapL(const TDesC& bitmapFile, TInt index); // Set background bitmap.
private: // from CCoeControl
    void Draw(const TRect& /*aRect*/) const;
private:
    CFont*	iFont;
    TDesC*  iText;
    CFbsBitmap* iBgndPicture;
};

//
// CNethackAppUi
//

class CNethackAppUi : public CEikAppUi
{
public:
    void ConstructL();
    ~CNethackAppUi();
    void CmdDialog();
    CNethackDocument* Document() const { return (CNethackDocument*)CEikAppUi::Document(); };
    const TFileName& configFile() { return iConfig; };
private: // from CEikAppUi
    void HandleCommandL(TInt aCommand);
    TBool ActivateIfRunning(const TDesC& ProgramName);
    TBool CNethackAppUi::FindFile(TDes& aFileName);
    void DimButton(TInt aCommand, TBool dimmed=ETrue);
    virtual void DynInitMenuPaneL(TInt aResourceId,CEikMenuPane* aMenuPane);
    void OpenDocL(TDesC& aFileName);
private:
    CNethackAppView* iAppView;
    TBool haveReadMe, haveManual, haveGuidebook, haveHints, haveGame, haveConfig;
    TFileName iReadMe;
    TFileName iManual;
    TFileName iGuidebook;
    TFileName iHints;
    TFileName iGame;
    TFileName iConfig;
public:
    TBuf<KMaxCommandLine> iArgs;
};

//
// CNethackDocument
//

class CNethackDocument : public CEikDocument, public MApaEmbeddedDocObserver
{
public:
    // construct/destruct
    CNethackDocument(CEikApplication& aApp);
    ~CNethackDocument();
    void ConstructL();
    static CNethackDocument* NewL(CEikApplication& aApp);

    inline void EditEmbeddedDocL(TInt aDocNum) {EmbeddedDoor(aDocNum)->DocumentL()->EditL(this);}
    inline CApaDoor* EmbeddedDoor(TInt aDocNum)const {return (CApaDoor*)((*iEmbedList)[aDocNum].iPicture.AsPtr());}
    //
    // from MApaEmbeddedDocObserver
    void NotifyExit(MApaEmbeddedDocObserver::TExitMode aMode);

    void EmbedExistingDocL(const TFullName& aFileName);
    void RemoveDocL(TInt index);
    void ViewDocL(const TDesC& aFilename);

    CArrayFixFlat<TPictureHeader>* iEmbedList;
private: // from CEikDocument
    CEikAppUi* CreateAppUiL();
};

//
// CNethackApplication
//

class CNethackApplication : public CEikApplication
{
private: // from CApaApplication
    CApaDocument* CreateDocumentL();
    TUid AppDllUid() const;
};

#endif
