// EIKCONCL.CPP
//
// Copyright (c) 1997-1999 Symbian Ltd.  All rights reserved.
//
#include "myconso.h"
//#include "eikconso.h"

#include <eikenv.h>
#include <eikappui.h>  
#include <eikapp.h>  
#include <eikkeys.h>
#include <basched.h>
#include <apgwgnam.h>
#include <eikcmds.hrh>
#include <fbs.h>
#include <coeutils.h>
#include <eikmenub.h>

#include "advconcl.h"
#include "advconcl.pan"
#include "NethackApp.rsg"

GLDEF_C void Panic(TEikConClPanic aPanic)
{
    User::Panic(_L("NH-CONCL"),aPanic);
}

enum
{
    EExit,
    ERead,
    EReadCancel,
    EWrite,
    ESetCursorPosAbs,
    ESetCursorPosRel,
    ESetCursorHeight,
    ESetTitle,
    EClearScreen,
    EClearToEndOfLine,
    ESetAttr,
    EKeyHit,
    ESetFont,
    EGetPointerPos
};

//
// class CAdvConsMessager
//

class CAdvConsMessager : public CActive
{
public:
    CAdvConsMessager(CMyConsoleScreen* aScreen,RThread aParentThread);
    ~CAdvConsMessager();
    void ConstructL(CAdvConsoleClient* aClient);
    void HandleKeyEvent(const TKeyEvent& aKeyEvent);
    void HandlePointerEvent(const TPointerEvent& aEvent, const TPoint& aPointer);
private: // overridden
    void RunL();
    void DoCancel();
private: // internal
    void CompleteReadRequest();
private:
    CMyConsoleScreen* iScreen;
    RThread iParentThread;
    TRequestStatus* iReadStatus;
    TKeyEvent* iKeyEvent;
    TInt iMessage;
    TPoint iPointer;
    const TAny* iParam;
    TRequestStatus* iReplyStatus;
    CCirBuf<TKeyEvent>* iKeyQ;
};

CAdvConsMessager::CAdvConsMessager(CMyConsoleScreen* aScreen,RThread aParentThread)
: CActive(EActivePriorityIpcEventsHigh),
iParentThread(aParentThread)
{
    iScreen=aScreen;
}

CAdvConsMessager::~CAdvConsMessager()
{
    iParentThread.Close();
    delete iKeyQ;
}

void CAdvConsMessager::ConstructL(CAdvConsoleClient* aClient)
{
    iKeyQ=new(ELeave) CCirBuf<TKeyEvent>;
    iKeyQ->SetLengthL(40);	// buffer length, too high? too low?
    iKeyEvent=(&aClient->iKeyEvent);
    aClient->iThreadStatus=(&iStatus);
    aClient->iMessage=(&iMessage);
    aClient->iParam=(&iParam);
    aClient->iReplyStatus=(&iReplyStatus);
    aClient->iScreen=iScreen;
    CActiveScheduler::Add(this);
    iStatus=KRequestPending;
    SetActive();
}

void CAdvConsMessager::DoCancel()
{
}

void CAdvConsMessager::RunL()
{
    switch (iMessage)
    {
    case EExit:
        CBaActiveScheduler::Exit();
        break;
    case ERead:
        if (iReadStatus)
            Panic(EEikConClPanicReadAlreadyOutstanding);

        // Bring the cursor onto the visible part of the screen whenever we read some input.
        iScreen->DrawCursorInSight();
        iReadStatus=(TRequestStatus*)iParam;
        if (iKeyQ->Count()>0) // already a buffered event
            CompleteReadRequest();
        break;
    case EReadCancel:
        if (iReadStatus)
            iParentThread.RequestComplete(iReadStatus,KErrCancel);
        break;
    case EWrite:
        iScreen->Write(*(const TDesC*)iParam);
        break;
    case ESetCursorPosAbs:
        iScreen->SetCursorPosAbs(*(const TPoint*)iParam);
        break;
    case ESetCursorPosRel:
        iScreen->SetCursorPosRel(*(const TPoint*)iParam);
        break;
    case ESetCursorHeight:
        iScreen->SetCursorHeight((TInt)iParam);
        break;
    case EClearScreen:
        iScreen->ClearScreen();
        break;
    case EClearToEndOfLine:
        iScreen->ClearToEndOfLine();
        break;
    case ESetAttr:
        iScreen->SetAtt((TInt)iParam);
        break;
    case EKeyHit:
    {
        // Returns the number of keys waiting in the buffer.
        TInt res = iKeyQ->Count();
        iParentThread.WriteL(iParam, TPckg<TInt>(res), 0);
        break;
    }
    case EGetPointerPos:
        iParentThread.WriteL(iParam, TPckgC<TPoint>(iPointer), 0);
        break;
    }
    iStatus=KRequestPending;
    SetActive();
    iParentThread.RequestComplete(iReplyStatus,0);
}

void CAdvConsMessager::CompleteReadRequest()
{
    if (iReadStatus)
    { 
        iKeyQ->Remove(iKeyEvent);;
        iParentThread.RequestComplete(iReadStatus,0);
    }
}

void CAdvConsMessager::HandleKeyEvent(const TKeyEvent& aKeyEvent)
{
    TInt ret=iKeyQ->Add(&aKeyEvent);
    if (ret==0)
        CEikonEnv::Beep();
    if (iKeyQ->Count()==1) // client may be waiting on this key event
        CompleteReadRequest();
}

// Handle pointer events by simulating a key event.
void CAdvConsMessager::HandlePointerEvent(const TPointerEvent& aEvent, const TPoint& aPointer)
{
    iPointer = aPointer;
    TKeyEvent dummy = { EKeyPointer, EKeyPointer, aEvent.iModifiers, 0 };
    HandleKeyEvent(dummy);
}

//
// class CAdvConsAppUi
//

struct SCommandLine
{
    RThread iParentThread;
    TRequestStatus* iStatus;
    CAdvConsoleClient* iClient;
    TSize iSize;
    const TDesC* iTitle;
    const TDesC* iFaceName;
    TInt iTwipSize;
};

const int NumFontSizes = 7;

class CAdvConsAppUi : public CEikAppUi
{
public:
    CAdvConsAppUi();
    void ConstructL(const SCommandLine* aComLine);
    ~CAdvConsAppUi();
private: // overridden
    void HandleKeyEventL(const TKeyEvent& aKeyEvent,TEventCode aType);
    void HandleForegroundEventL(TBool aForeground);
    void SetAndDrawFocus(TBool aFocus);
    TBool FindFile(TDes& aFileName, const TDesC *aDefault = NULL);
    void AddMenu(void);
private:
    TInt iZoomFactor;
    TBuf<20> iFaceName;
    TInt iTwipSize;

    void SetZoomFactorL(TInt aZoomFactor);
    void ZoomOutL();
    void ZoomInL();
    void HandleCommandL(TInt aCommand);
    void HandlePointerEventL(const TPointerEvent& aPointerEvent);
    void HandleWsEventL(const TWsEvent& aEvent,CCoeControl* aDestination);
    TPoint PixelsToChars(TPoint aPos);
    void CalculateInitialFontSizeL(TInt aMin, TInt aMax);

    CMyConsoleScreen* iScreen;
    CMyConsoleControl* iControl;
    CAdvConsMessager* iMessager;
    TInt iFontId;
    CFbsTypefaceStore *iFontStore;
    struct CFontInfo {
        TInt iTwips;    // Height of the font in Twips.
        TSize iSize;    // Size of the resulting window in characters.
        CFontInfo(): iTwips(0), iSize() {};
    } iFontInfo[NumFontSizes];
    TInt iNumFonts;
    TSize iConsoleSize; // Screen size in pixels.
    // TInt iFontSizes[NumFontSizes];
};

CAdvConsAppUi::CAdvConsAppUi():
   iFontId(-1), iFontStore(NULL)
{
}

CAdvConsAppUi::~CAdvConsAppUi()
{
    delete(iScreen);
    delete(iMessager);
    if (iFontStore)
    {
        iFontStore->RemoveFile(iFontId);
        delete iFontStore;
    }
}

TBool CAdvConsAppUi::FindFile(TDes& aFileName, const TDesC *aDefault)
{
    TParse parser;
#if defined(__WINS__)
	TFileName appName = _L("Z:\\System\\Apps\\NethackApp\\Nethack.DXE");
#else
    TFileName appName = RProcess().FileName();
#endif
    User::LeaveIfError(parser.SetNoWild(aFileName, aDefault, &appName));

    TPtrC theFile (parser.FullName());
    if (! ConeUtils::FileExists(theFile)) 
        return EFalse;
    aFileName = parser.FullName();
    return ETrue;
}

void CAdvConsAppUi::ConstructL(const SCommandLine* aComLine)
{
    iFaceName = *aComLine->iFaceName;
    iTwipSize = aComLine->iTwipSize;
    
    CEikAppUi::BaseConstructL(ENoAppResourceFile);

    // Loads an array of font sizes. Also sets iZoomFactor to the
    // initial zoom.
    CalculateInitialFontSizeL(5*20, 18*20);

    iScreen=new(ELeave) CMyConsoleScreen;
    iScreen->ConstructL(*(aComLine->iTitle), 0, TFontSpec(iFaceName, iFontInfo[iZoomFactor].iTwips));
    iControl=iScreen->ConsoleControl();

    iScreen->SetKeepCursorInSight(EFalse);
    iScreen->SetScrollBarVisibilityL(CEikScrollBarFrame::EAuto,CEikScrollBarFrame::EAuto);

    iControl->SetFocus(ETrue,EDrawNow);
    iMessager=new(ELeave) CAdvConsMessager(iScreen,aComLine->iParentThread);
    iMessager->ConstructL(aComLine->iClient);
    RThread().SetPriority(EPriorityMore);

    AddMenu();
}

// Calculate the font sizes to use. We basically want an 80x25 screen, starting with the largest font that
// gives us this size (or the smallest font if none quite manages it).
// Zoom factors are then set as seems appropriate.
void CAdvConsAppUi::CalculateInitialFontSizeL(TInt aMin, TInt aMax)
{
    TSize lastSize(0,0);
    TInt zoom = 0;
    iConsoleSize = iCoeEnv->ScreenDevice()->SizeInPixels();

    // Try to load the extra font file.
    TFileName fontFile = iFaceName;
    TPtrC deflt = _L("*.gdr");
    if (FindFile(fontFile, &deflt))
    {
        // Try to load the extra font file!
        iFontStore = CFbsTypefaceStore::NewL(iCoeEnv->ScreenDevice());
        iFontStore->AddFile(fontFile, iFontId);
    }

    for (TInt twipSize = aMin; twipSize <= aMax && zoom < NumFontSizes; twipSize+=10)
    {
        CFbsFont* font=iEikonEnv->CreateScreenFontL(TFontSpec(iFaceName, twipSize));
        // Work out whether this font is suitable
        TInt width = iConsoleSize.iWidth/font->MaxNormalCharWidthInPixels();
        TInt height = iConsoleSize.iHeight/font->HeightInPixels();
        iEikonEnv->ReleaseScreenFont(CONST_CAST(CFont*, (CFont*)font)); // Dont need font any more.

        TSize size(width, height);
        
        // Ignore duplicate fonts --- we didn't increase the size
        // enough to get a new one. 
        if (size == lastSize) continue;
        
        if (width >= 80 && height >= 25)
        {
            // Pick the largest font that fits as the default zoom
            iZoomFactor = zoom;
        }

        iFontInfo[zoom].iTwips = twipSize;
        iFontInfo[zoom].iSize = size;
        zoom++;
        lastSize = size;
    }
    iNumFonts = zoom;    // Total number of fonts.
}

void CAdvConsAppUi::HandleForegroundEventL(TBool aForeground)
{
    if (aForeground)
        RThread().SetPriority(EPriorityMore);
}

void CAdvConsAppUi::SetAndDrawFocus(TBool aFocus)
{
    if (iControl)
        iControl->SetFocus(aFocus,EDrawNow);
}

void CAdvConsAppUi::HandleKeyEventL(const TKeyEvent& aKeyEvent,TEventCode aType)
{
    if (aType==EEventKey) switch(aKeyEvent.iCode) {
    case EEikSidebarZoomInKey:
        ZoomInL();
        break;
    case EEikSidebarZoomOutKey:
        ZoomOutL();
        break;
    default:
        iMessager->HandleKeyEvent(aKeyEvent);
    }
}

//
// class CConsEikonEnv
//

class CConsEikonEnv : public CEikonEnv
{
public:
    void ConstructConsoleEnvironmentL(const SCommandLine* aComLine);
};

void CConsEikonEnv::ConstructConsoleEnvironmentL(const SCommandLine* aComLine)
{
    ConstructL();
    CAdvConsAppUi* appUi=new(ELeave) CAdvConsAppUi;
    appUi->ConstructL(aComLine);
    CApaWindowGroupName* wgName=CApaWindowGroupName::NewLC(iWsSession);
    TPtrC caption=*(aComLine->iTitle);
    wgName->SetCaptionL(caption);
    wgName->SetRespondsToShutdownEvent(EFalse);
    wgName->SetRespondsToSwitchFilesEvent(EFalse);
    wgName->SetWindowGroupName(iRootWin);
    CleanupStack::PopAndDestroy(); // wgName
#if defined(__EPOC32__)
    RProcess().Rename(caption);
#endif
    RThread().Rename(caption);
}

TInt ConsoleClientStartFunction(TAny* aParam)
{
    const SCommandLine* comLine=(const SCommandLine*)aParam;
    TInt err=KErrNoMemory;
    CConsEikonEnv* coe=new CConsEikonEnv;
    if (coe)
        TRAP(err,coe->ConstructConsoleEnvironmentL(comLine));
    TRequestStatus* pS=(comLine->iStatus);
    comLine->iParentThread.RequestComplete(pS,err);
    if (!err)
        coe->ExecuteD();
    return(0);
}

//
// class CAdvConsoleClient
//

CAdvConsoleClient::~CAdvConsoleClient()
{
    if (iLogonStatus.Int()==KRequestPending && iReplyStatus)
        SendReceive(EExit,NULL);
    iThread.Close();
}

CAdvConsoleClient::CAdvConsoleClient(const TDesC *aFaceName)
{
    if (aFaceName==NULL)
        iFaceName = _L("Nethack");
    else
        iFaceName = *aFaceName;
    iTwipSize = 220;
}

const TInt KMaxHeapSize=0x1000*254; // chunks are a megabyte anyway

TInt CAdvConsoleClient::Create(const TDesC& aTitle,TSize aSize)
{ 
    TInt err;
    TRequestStatus status=KRequestPending;
    SCommandLine comLine;
    comLine.iStatus=(&status);
    comLine.iClient=this;
    comLine.iSize=aSize;
    comLine.iTitle=&aTitle;
    comLine.iFaceName = &iFaceName;
    comLine.iTwipSize = iTwipSize;
    TBuf<20> threadName;
    TInt num=0;
    do
    {
        threadName.Format(_L("UI%02d"),num++); // !! review the title
        err=iThread.Create(threadName,ConsoleClientStartFunction,KDefaultStackSize,KMinHeapSize,KMaxHeapSize,&comLine,EOwnerThread);
    } while(err==KErrAlreadyExists);
    if (!err)
    {
        iThread.Logon(iLogonStatus);
        comLine.iParentThread.Duplicate(iThread);
        iThread.Resume();
        User::WaitForRequest(status,iLogonStatus);
        err=status.Int();
    }
    return(err);
}

void CAdvConsoleClient::SendReceive(TInt aMessage,const TAny* aParam)
{
    if (iLogonStatus.Int()!=KRequestPending)
        User::Exit(KErrCancel);
    *iMessage=aMessage;
    *iParam=aParam;
    TRequestStatus replyStatus=KRequestPending;
    *iReplyStatus=(&replyStatus);
    TRequestStatus* pS=iThreadStatus;
    iThread.RequestComplete(pS,0);
    User::WaitForRequest(replyStatus,iLogonStatus);
}

void CAdvConsoleClient::Read(TRequestStatus& aStatus)
{
    aStatus=KRequestPending;
    SendReceive(ERead,&aStatus);
}

void CAdvConsoleClient::ReadCancel()
{
    SendReceive(EReadCancel,NULL);
}

void CAdvConsoleClient::Write(const TDesC& aDes)
{
    SendReceive(EWrite,&aDes);
}

TPoint CAdvConsoleClient::CursorPos() const
{
    return(iScreen->CursorPos());
}

TPoint CAdvConsoleClient::PointerPos()
{
    TPoint position;
    TPckg<TPoint> p(position);
    SendReceive(EGetPointerPos, &p);
    return position;
}

void CAdvConsoleClient::SetCursorPosAbs(const TPoint& aPosition)
{
    SendReceive(ESetCursorPosAbs,&aPosition);
}

void CAdvConsoleClient::SetCursorPosRel(const TPoint &aVector)
{
    SendReceive(ESetCursorPosRel,&aVector);
}

void CAdvConsoleClient::SetCursorHeight(TInt aPercentage)
{
    SendReceive(ESetCursorHeight,aPercentage);
}

void CAdvConsoleClient::SetTitle(const TDesC& aTitle)
{
    SendReceive(ESetTitle,&aTitle);
}

void CAdvConsoleClient::ClearScreen()
{
    SendReceive(EClearScreen,NULL);
}

void CAdvConsoleClient::ClearToEndOfLine()
{
    SendReceive(EClearToEndOfLine,NULL);
}

TSize CAdvConsoleClient::ScreenSize() const
{
    return(iScreen->ScreenSize());
}

TKeyCode CAdvConsoleClient::KeyCode() const
{
    return((TKeyCode)iKeyEvent.iCode);
}

TUint CAdvConsoleClient::KeyModifiers() const
{
    return(iKeyEvent.iModifiers);
}

TInt CAdvConsoleClient::KeyHit() 
{
    TInt Status = 0;
    TPckg<TInt> p(Status);
    SendReceive(EKeyHit, &p);
    return Status;
}

// Call this before constructing to set the default font.
void CAdvConsoleClient::SetFont(TDesC *aFaceName, TInt aPointSize)
{
    iFaceName = *aFaceName;
    iTwipSize = aPointSize*20;
    // If we have initialised, reset the screen.
}

CAdvConsoleClient* NewAdvConsole(const TDesC *aFaceName)
{
#if defined(__WINS__)
    // return null if the graphical window server thread is not running
    TFindThread findT(_L("Wserv"));
    TFullName name;
    if (findT.Next(name)!=KErrNone)
        return(NULL);
#endif
    return(new CAdvConsoleClient(aFaceName));
}

void CAdvConsoleClient::SetAttr(int attrib)
{
    SendReceive(ESetAttr,attrib);
}


void CAdvConsoleClient::Color(TUint /*aForegroundGrey16*/, TUint /*aBackgroundGrey16*/)
{
    // Reserved for future use.
}

void CAdvConsAppUi::HandleCommandL(TInt aCommand)
{
    // Handle the command generated by:
    //   1. menu item selection
    //   2. short-cut key press
    //   3. tool bar item selection
    switch (aCommand)
    {
        // Zoom in command
    case EEikCmdZoomIn:             // <---Pre-defined command id
        ZoomInL();
        break;
        // Zoom out command
    case EEikCmdZoomOut:            // <---Pre-defined command id
        ZoomOutL();
        break;
    default:    // Menu command: the command ID should be the same as the key press.
        while (aCommand != 0)
        {
            TInt code = aCommand & 0x7f;
            TInt modifier = 0;
            if (aCommand & 0x80) modifier |= EModifierFunc|EModifierAlt;
            if (code < 0x20) modifier |= EModifierCtrl;
            TKeyEvent dummy = { code, code, modifier, 0 };
            iMessager->HandleKeyEvent(dummy);
            aCommand >>= 8;
        }
        break;
    }
}

void CAdvConsAppUi::ZoomInL()
{
    TInt zoom = iZoomFactor+1;
    if (zoom >= iNumFonts) zoom = 0;
    SetZoomFactorL(zoom);
}

void CAdvConsAppUi::ZoomOutL()
{
    TInt zoom = iZoomFactor - 1;
    if (zoom < 0) zoom = iNumFonts - 1;
    SetZoomFactorL(zoom);
}

void CAdvConsAppUi::SetZoomFactorL(TInt aZoomFactor)
{
    iZoomFactor = aZoomFactor;
    iScreen->SetFontL(TFontSpec(iFaceName, iFontInfo[iZoomFactor].iTwips));

#if 0
    TBuf<256> buffer;
    buffer.Format(_L("Font %d,%d,%d(%dx%d), Screen size %d x %d (%dx%d),(%dx%d)"),
                  iFontInfo[iZoomFactor].iTwips,
                  iScreen->ConsoleControl()->CharSize().iWidth,
                  iScreen->ConsoleControl()->CharSize().iHeight,
                  iFontInfo[iZoomFactor].iSize.iWidth,
                  iFontInfo[iZoomFactor].iSize.iHeight,
                  iScreen->ConsoleControl()->Size().iWidth/iScreen->ConsoleControl()->CharSize().iWidth,
                  iScreen->ConsoleControl()->Size().iHeight/iScreen->ConsoleControl()->CharSize().iHeight,
                  iScreen->ConsoleControl()->Size().iWidth,
                  iScreen->ConsoleControl()->Size().iHeight,
                  iScreen->ScreenSize().iWidth,
                  iScreen->ScreenSize().iHeight);

    CEikonEnv::Static()->InfoMsg(buffer);
#endif
}

void CAdvConsAppUi::HandlePointerEventL(const TPointerEvent& /*aPointerEvent*/)
{
#if 0
    TBuf<128> iMessage;
    iMessage.Format(_L("Pointer event %d, (%d,%d)"),
                    aPointerEvent.iType, aPointerEvent.iPosition.iX, aPointerEvent.iPosition.iY);
    CEikonEnv::Static()->InfoMsg(iMessage);
#endif
}

TPoint CAdvConsAppUi::PixelsToChars(TPoint aPos)
{
    return TPoint(aPos.iX / iControl->CharSize().iWidth, aPos.iY / iControl->CharSize().iHeight);
}

void CAdvConsAppUi::HandleWsEventL(const TWsEvent& aEvent,CCoeControl* aDestination)
{
    if (aEvent.Type() == EEventPointer && aDestination == iControl)
    {
        TPointerEvent aPointerEvent = *aEvent.Pointer();
        TPoint charPos = PixelsToChars(aPointerEvent.iPosition);
        // This is why I need the disgusting kludge above.
//        charPos += iScreen->TopLeft();   // Handle the case where the window is scrolled.

#if 0
        TBuf<128> iMessage;
        iMessage.Format(_L("Pointer event %d, (%d,%d)"),
                        aPointerEvent.iType, charPos.iX, charPos.iY);
        CEikonEnv::Static()->InfoMsg(iMessage);
#endif
        switch(aPointerEvent.iType)
        {
        case TPointerEvent::EButton1Down:
            iMessager->HandlePointerEvent(aPointerEvent, charPos);
            break;
//        case TPointerEvent::EButton1Up:
//            iControl->Window().RequestPointerRepeatEvent(KPointerRepeatRate,rect);
//            break;
        default: break;
        }


    }
    CEikAppUi::HandleWsEventL(aEvent, aDestination);
}

// Add a menu for nethack commands.
void CAdvConsAppUi::AddMenu()
{
    TInt resourceFileOffset = -1;

    TFileName resourceFile;
#if defined(__WINS__)
    TPtrC aDefault = _L("nethackapp.rsc");
#else
    TPtrC aDefault = _L("nethack.rsc");
#endif
    if (FindFile(resourceFile, &aDefault))
    {
        resourceFileOffset = iEikonEnv->AddResourceFileL(resourceFile);
        iMenuBar = new (ELeave) CEikMenuBar;
        iMenuBar->ConstructL(this, R_NETHACK_HOTKEYS, R_NETHACK_MENUBAR);
        AddToStackL(iMenuBar,ECoeStackPriorityMenu,ECoeStackFlagRefusesFocus);
    }
}
