// EIKCONSO.H
//
// Copyright (c) 1997-1999 Symbian Ltd.  All rights reserved.
//

#if !defined(__MYCONSO_H__)
#define __MYCONSO_H__

#if !defined(__EIKCONSO_H__)
#include <eikconso.h>
#endif

class CMyConsoleControl;
class CMyConsoleScreen : public CConsoleBase
	{
public:
	enum TEikConsoleFlags
		{
		EUseBackedUpWindow	=0x1,
		ENoInitialCursor	=0x2
		};
public:
	CMyConsoleScreen();
	~CMyConsoleScreen();
	TInt Create(const TDesC& aTitle,TSize aSize);
	void Read(TRequestStatus& aStatus);
	void ReadCancel();
	TKeyCode KeyCode() const;
	TUint KeyModifiers() const;
	void ConstructL(const TDesC &aTitle,TInt aFlags, const TFontSpec& aFontSpec);// use full screen 
	void ConstructL(const TDesC &aTitle,const TSize &aSize,TInt aFlags,TEikConsWinUnits aUnit=EEikConsWinInChars); // place the screen at TPoint(0,0)
	void ConstructL(const TDesC &aTitle,TPoint aTopLeft,const TSize &aSize,TInt aFlags,TEikConsWinUnits aUnit=EEikConsWinInChars); 
	void SetConsWin(CMyConsoleControl *aConsWin); // use if you don't want WHOLE screen
	inline CMyConsoleControl* ConsoleControl() const {return iConsoleControl;};

	void SetTitle(const TDesC &aTitle);
	void SetTitleL(const TDesC &aTitle);
	inline HBufC *Title() const {return iConsoleTitle;};

	TSize ScreenSize() const;

	void FlushChars(); //new
	TPoint CursorPos() const;
	void SetCursorPosAbs(const TPoint &aPoint);		
	void SetCursorPosRel(const TPoint &aPoint);		
	void SetCursorHeight(TInt aPercentage);		
	void Write(const TDesC &aDes);				   		
	
	void ClearScreen();				
	void ClearToEndOfLine();		
	void ClearChars(const TRect &aRect,TUint aCharacterAttributes);	

	TInt HideCursor(); // returns nonzero if cursor wasn't visible
	void DrawCursor(); 
	void DrawInSight(TPoint aPosition);
	void DrawCursorInSight();
	
	void Up(TUint aCount=1);		
	void Down(TUint aCount=1);	
	void Right(TUint aCount=1);	
	void Left(TUint aCount=1);	
	void Cr();							
	void Lf();							

	void ScrollChars(const TRect &anArea,const TPoint &aVector);
	void Redraw(const TRect &aRect);	// called by CMyConsoleControl
	void MoveTopLeft(TPoint aVector);	// called by CMyConsoleControl
        TPoint TopLeft() { return iTopLeft; };  // Make top,left accessible. DGB

	inline TUint HistorySize() const {return iNoChrExtra;}; // lines ACTUALLY stored				
	void SetHistorySizeL(TUint aMaxChrExtra,TUint aMaxAttExtra);  // lines of back-scroll history resp. lines of ATTRIBUTED backscroll history (aMaxAttExtra<=aMaxChrExtra)
		
	inline TUint Att() const {return iAtt;};
	
	void SetAtt(TUint aCharacterAttributes);		
	void SetAtt(TUint aForegroundColor16,TUint aBackgroundColor16); 

	void SetFontL(const TFontSpec &aFontDesc);	
	const TFontSpec& Font() const;

	void SetSelection(const TRect &aRange); 
	void SelectCursor(); // empty selection starting and ending at cursor position
	TRect Selection();
	HBufC *RetrieveL(const TRect &aRange);	// returns selection in newly allocated buffer

	void SetPureCRLF(TUint aFlag);
	void SetAllPrintable(TUint aFlag);
	void SetScrollLock(TUint aFlag);
	void SetKeepCursorInSight(TUint aFlag);
	void SetScrollBarVisibilityL(CEikScrollBarFrame::TScrollBarVisibility aHBarVisibility, CEikScrollBarFrame::TScrollBarVisibility aVBarVisibility);
	TBool UpdateScrollBars();

private:
	// high-level history manipulation
	void MemScrPut(const TDesC &aString,TPoint aPos,TUint8 aCharacterAttributes);
	void MemScrClr(const TRect &aRect,TUint8 aCharacterAttributes);
	void MemScrScrollUp(TUint aLines=1);
	void MemScrMove(TPoint anOldPt,TPoint aNewPt,TUint aLen);
	void MemScrScroll(const TRect &aRect,TPoint aVector);
	// low-level history access
	TPtr	MemChrPtr(TPoint aPos,TUint aLen);
	TPtr8	MemAttPtr(TPoint aPos,TUint aLen);
	// console screen-to-visible window conversion
	TPoint 	ScrToVis(TPoint aPoint);
	TRect 	ScrToVis(const TRect &aRect);
	// recalculates iVisSize and iVisWin
	TBool RecalculateSize();	
	// calculate a TopLeft that would put aPosition in-sight
	TPoint GetAnInSightPosition(TPoint aPosition) const;
	// draw cursor or draw cursor in sight, depending on iKeepCursorInSight
	void DrawCursorWhereNeeded();
	// smart cursor-clipping: clips cursor to within screen, but allows it to be past the last character
	void ClipCursor();
	// redraw characters with doublewidth LEFT/RIGHT flag set depending on odd/even columns
	void CMyConsoleScreen::RedrawChars(TInt anX,TInt anY,TInt aLength,TUint aCharacterAttributes);
	// routines needed for selection
	void InvertOverlap(TPoint aStart,TPoint aEnd,TPoint bStart,TPoint bEnd);
	void InvertRange(const TRect &aRect);

private:
	CMyConsoleControl *iConsoleControl;// console window (handles redraw events)
	TSize iConsoleSize; 		// size of console in characters (e.g. 80x24)	
	HBufC *iConsoleTitle;		// console title
	TRect iSelection; 			// current selection
	TPoint iCursor;				// current cursor position
	TUint iAtt;					// current character attribute

	TInt iWriteNow;				// if FALSE, Write()-calls can be buffered
	TBuf<256> iWriteBuffer;		// the Write() buffer
	TInt	iMaxChrExtra;		// max nr of extra lines of characters stored
	TInt	iNoChrExtra;		// actual nr of extra lines of characters stored
	TInt	iMaxAttExtra;		// max nr of extra lines of character attributes stored
	TText	*iChrBuf;			// character history
	TUint8	*iAttBuf;			// attribute history
	TPoint	iTopLeft;			// coordinates of top left corner of visual display

	TRect iVisWin;				// RECT visible window
	TSize iVisSize;				// SIZE visible window

	// settings
	TUint iPureCRLF;			// FALSE by default. If TRUE, CR and LF are "pure"
	TUint iAllPrintable;		// FALSE by default. If TRUE, even chars 0-31 are printed
	TUint iScrollLock;			// FALSE by default. If TRUE, will not scroll at bottom line
	TUint iKeepCursorInSight;	// TRUE by default. If FALSE, cursor may be offscreen after a Write()
	TUint iCursorBlinking; 		// TRUE if cursor should be visible (i.e. user WANTS it to be visible)

	TInt iConsFlags;

	};

class CMyConsoleControl : public CCoeControl, public MEikScrollBarObserver 
	{
public:	
	// constructors, destructors, initialization
	CMyConsoleControl();
	~CMyConsoleControl();
	void	ConstructL(TInt aFlags, const TFontSpec& aFontSpec);
	void	ConstructL(TPoint aTopLeft,const TSize &aSize,TInt aFlags,TEikConsWinUnits aUnit);
	void 	SetRedrawer(CMyConsoleScreen *aConsole);

	// font stuff
	void	SetFontL(const TFontSpec &aFontSpec,const CFbsFont* aFont);		// change to font aFontSpec
	inline const TFontSpec& Font() const {return iFontSpec;};	// returns current fontspec

	inline const TSize& CharSize() const {return iCharSize;};	
	TSize	VisibleSize() const;						// returns maximal visible nr of visible characters

	// cursor stuff
	void	SetCursorHeight(TUint aPercentage);	// set cursor height (0-100%)
	void	DrawCursor(TPoint aPosition);		// draw cursor at character position
	void	HideCursor();						// hide cursor

	// basic graphic functions
	void	DrawChars(const TDesC &aString,const TPoint &aPosition,TUint aCharacterAttributes);	// draw aString at character position aPosition using
	void	ClearChars(const TRect &anArea,TUint aCharacterAttributes);	// clear the character area
	void	InvertChars(const TRect &anArea);	// invert the character area
	void	ScrollChars(const TRect &anArea,const TPoint &aVector);	// scroll characters inside anArea by the given vector

	// scrollbar stuff
	TInt	SetScrollBarVisibilityL(CEikScrollBarFrame::TScrollBarVisibility aHBarVisibility, CEikScrollBarFrame::TScrollBarVisibility aVBarVisibility);
	TBool	UpdateScrollbarsL(const TSize &aDataSize,const TSize& aVisibleSize,TPoint aPos);
	void	HandleScrollEventL(CEikScrollBar* aScrollBar,TEikScrollEvent aEventType);
	void	UpdateArea();
        TBool ColorDisplay() { return iColorDisplay; };


protected:	// virtual
	// window stuff
	void	Draw(const TRect& aRect) const;	
	void	SizeChangedL();
	void FocusChanged(TDrawNow aDrawNow);
	virtual TInt CountComponentControls() const;
	virtual CCoeControl* ComponentControl(TInt aIndex) const;

private: // virtual - Reserved from CCoeControl
	virtual void Reserved_1();
	virtual void Reserved_2();

private:
	TPoint	CharsToPixels(TPoint aPoint);
	TRect	CharsToPixels(const TRect &aRect);
	TRect	PixelsToChars(const TRect &aRect);
	void	ClearPixels(const TRect &anArea,TUint aColor16) const;

	void 	NextFontL(); 			
	void	InterpretAttribute(TUint AnAtt);
	void	InterpretColorBits(TUint aCharacterAttributes);
	inline 	CMyConsoleControl &M() const {return((CMyConsoleControl&)(*this));}
 	void CreateScrollBarFrameLayout(TEikScrollBarFrameLayout& aLayout) const;

private:
	CMyConsoleScreen *iConsole;	// the console that owns this window
	TUint iBackColor16;				// current background color 
	TUint iPenColor16;				// current pen color for characters
	TUint iLastAtt;					// last attribute, to see if we need to change the gc
	TUint iLastFontFlags;			// last attribute, to see if we need to load another font

	TSize iCharSize;				// size of a single character (in pixels)
	TSize iViewInPixels;			// size of area available for characters (in pixels)
	TSize iViewInChars;				// size of area available for characters (in characters)

	TTextCursor iTextCursor;		// cursor to use
	TUint iCursorHeightPercentage;	// required height of cursor (as a percentage)

	TInt iRedrawing;				// NONZERO IF BUSY REDRAWING
	
	CEikScrollBarFrame* iSBFrame;
	CEikScrollBarFrame::TScrollBarVisibility	iHBarVisibility;
	CEikScrollBarFrame::TScrollBarVisibility	iVBarVisibility;
	TPoint	iLastThumbPos;

	const CFont *iCurrentFont;		// Current font
	TFontSpec iFontSpec;			// Current font spec
	TFontUnderline iFontUnderline;	// Current font spec underline
	TUint iFontHeight;				// Height of current font when not doubleheight/doublewidth
	TUint iFontIsProportional; 		// nonzero if font is proportional
        TBool iColorDisplay;    // TRUE if the display supports colour.
	};

#endif
