// My TTY alternative to stdin, stdout, stderr supplied by Psion stdlib implementation.
// This version is very similar, but supports some additional functions, such as moving the
// cursor around the window, or setting the window title.
#include <e32std.h>
#include <e32cons.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/unistd.h>
#include <ctype.h>
#include "advconcl.h"
#include "mytty.h"
extern "C" {
#pragma warning(disable:4510 4610)
#include "hack.h"
#include "color.h"
#include "wintty.h"
}

extern unsigned int EpocColorDisplay;

int cursor_x, cursor_y, cursor_click;
int last_colour = NO_COLOR;

#if !defined(CTRL)
#define CTRL(x) ((x)-'a'+1)
#endif

#if defined(__WINS__) && defined(__EXE__)
#define WIDTHKLUDGE 2
#define HEIGHTKLUDGE 2
#else
#define WIDTHKLUDGE 0
#define HEIGHTKLUDGE 0
#endif

// Epoc 16 colour mode
enum EPOC16COLOURS {
    EC_BLACK,
    EC_DARKGREY,
    EC_RED,
    EC_BROWN,
    EC_GREEN,
    EC_BRIGHTRED,
    EC_YELLOW,
    EC_BRIGHTGREEN,
    EC_BRIGHTMAGENTA,
    EC_BRIGHTBLUE,
    EC_BRIGHTCYAN,
    EC_MAGENTA,
    EC_BLUE,
    EC_CYAN,
    EC_GREY,
    EC_WHITE
};

// Map nethack colours to Epoc colours.
static unsigned char ColorTable[] = {
    ATT_COLORMASK | EC_DARKGREY, // Black (nearly!)
    ATT_COLORMASK | EC_RED, // Red
    ATT_COLORMASK | EC_GREEN, // Green
    ATT_COLORMASK | EC_BROWN, // Brown
    ATT_COLORMASK | EC_BRIGHTBLUE, // Blue
    ATT_COLORMASK | EC_MAGENTA, // Magenta
    ATT_COLORMASK | EC_CYAN, // Cyan
    ATT_COLORMASK | EC_GREY, // Gray
    ATT_COLORMASK | EC_GREY, // No color
    ATT_COLORMASK | EC_BRIGHTRED, // Orange
    ATT_COLORMASK | EC_BRIGHTGREEN, // Bright green
    ATT_COLORMASK | EC_YELLOW, // Yellow
    ATT_COLORMASK | EC_BRIGHTBLUE, // Bright blue
    ATT_COLORMASK | EC_BRIGHTMAGENTA, // Bright Magenta
    ATT_COLORMASK | EC_BRIGHTCYAN, // Bright Cyan
    ATT_COLORMASK | EC_WHITE, // White.
};

static unsigned char InverseColorTable[] = {
    ATT_COLORMASK | 0x70 | EC_DARKGREY, // Black (nearly!)
    ATT_COLORMASK | 0x70 | EC_RED, // Red
    ATT_COLORMASK | 0x70 | EC_GREEN, // Green
    ATT_COLORMASK | 0x70 | EC_BROWN, // Brown
    ATT_COLORMASK | 0x70 | EC_BLUE, // Blue
    ATT_COLORMASK | 0x70 | EC_MAGENTA, // Magenta
    ATT_COLORMASK | 0x70 | EC_CYAN, // Cyan
    ATT_COLORMASK | 0x70 | EC_GREY, // Gray
    ATT_COLORMASK | 0x70 | EC_GREY, // No color
    ATT_COLORMASK | 0x70 | EC_BRIGHTRED, // Orange
    ATT_COLORMASK | 0x70 | EC_BRIGHTGREEN, // Bright green
    ATT_COLORMASK | 0x70 | EC_YELLOW, // Yellow
    ATT_COLORMASK | 0x70 | EC_BRIGHTBLUE, // Bright blue
    ATT_COLORMASK | 0x70 | EC_BRIGHTMAGENTA, // Bright Magenta
    ATT_COLORMASK | 0x70 | EC_BRIGHTCYAN, // Bright Cyan
    ATT_COLORMASK | 0x70 | EC_BLACK, // White.
};

// Inverted colours for mono displays:
#define EM_WHITE        EC_BLACK | 0x70
#define EM_GREY         EC_DARKGREY | 0x70
#define EM_DARKGREY     EC_GREY | 0x70

static unsigned char MonoTable[] = {
    ATT_COLORMASK | EM_WHITE, // Black - shows as white on a black background.
    ATT_COLORMASK | EM_DARKGREY, // Red
    ATT_COLORMASK | EM_GREY, // Green
    ATT_COLORMASK | EM_DARKGREY, // Brown
    ATT_COLORMASK | EM_DARKGREY, // Blue
    ATT_COLORMASK | EM_GREY, // Magenta
    ATT_COLORMASK | EM_GREY, // Cyan
    ATT_COLORMASK | EM_GREY, // Gray
    ATT_COLORMASK | EM_GREY, // No color
    ATT_COLORMASK | EM_GREY, // Orange
    ATT_COLORMASK | EM_WHITE, // Bright green
    ATT_COLORMASK | EM_GREY, // Yellow
    ATT_COLORMASK | EM_GREY, // Bright blue
    ATT_COLORMASK | EM_WHITE, // Bright Magenta
    ATT_COLORMASK | EM_WHITE, // Bright Cyan
    ATT_COLORMASK | EM_WHITE, // White.
};

static unsigned char InverseMonoTable[] = {
    ATT_COLORMASK | EC_WHITE, // Black - shows as white on a black background.
    ATT_COLORMASK | EC_DARKGREY, // Red
    ATT_COLORMASK | EC_GREY, // Green
    ATT_COLORMASK | EC_DARKGREY, // Brown
    ATT_COLORMASK | EC_DARKGREY, // Blue
    ATT_COLORMASK | EC_GREY, // Magenta
    ATT_COLORMASK | EC_GREY, // Cyan
    ATT_COLORMASK | EC_GREY, // Gray
    ATT_COLORMASK | EC_GREY, // No color
    ATT_COLORMASK | EC_GREY, // Orange
    ATT_COLORMASK | EC_WHITE, // Bright green
    ATT_COLORMASK | EC_GREY, // Yellow
    ATT_COLORMASK | EC_GREY, // Bright blue
    ATT_COLORMASK | EC_WHITE, // Bright Magenta
    ATT_COLORMASK | EC_WHITE, // Bright Cyan
    ATT_COLORMASK | EC_WHITE, // White.
};

static void MagicFile(FILE *f, void* _cookie,
                  int	(*readfn) (void* _cookie, char *_buf, int _n),
                  int	(*writefn)(void* _cookie, const char *_buf, int _n),
                  fpos_t	(*seekfn) (void* _cookie, fpos_t _off, int _whence),
                  int	(*closefn)(void* _cookie))
{
    if (f == NULL) return;
    memset(f, sizeof(f), 0);
    f->_cookie = _cookie;
    f->_read = readfn;
    f->_write = writefn;
    f->_seek = seekfn;
    f->_close = closefn;
    f->_data = ImpurePtr();
}


#define	fropen(cookie, fn) funopen(cookie, fn, (int (*)())0, (fpos_t (*)())0, (int (*)())0)
#define	fwopen(cookie, fn) funopen(cookie, (int (*)())0, fn, (fpos_t (*)())0, (int (*)())0)

/*	CLineEdit class - borrowed from F32 text window server */

class CLineEdit : public CBase
{
public:
    enum TCursorType {ECursorNone=0, ECursorNormal=100, ECursorInsert=20};
    enum TEditMode {EEditOverWrite,EEditInsert};
public:
    static CLineEdit* NewL(CAdvConsoleClient* aConsole,TInt aMaxHistory);
    ~CLineEdit();
    void Edit(const TDesC& aPrompt,TDes* aBuf);
protected:
    CLineEdit();
    TInt ApparentPos(int aPos);
    TPoint Where(int aPos = -1);
    TInt Lines();
    TInt WordLeft();
    TInt WordRight();
    void ClearLine();
    void ClearLast(TInt aCnt);
    void Recall();
    void Cursor();
    void Refresh();
    inline TDes& Buf() {return *iBuf;}
private:
    CArrayFixFlat<HBufC*>* iHistory;
    CAdvConsoleClient* iConsole; // Not owned
    TInt iMaxHistory;
    TInt iWidth;
    TInt iHeight;
    TInt iPos;
    TInt iLine;
    TInt iOrigin;
    TInt iRecall;
    TEditMode iMode;
    TDes* iBuf;
};

CLineEdit::CLineEdit()
//
// Constructor
//
{

//	iHistory=NULL;
//	iConsole=NULL;
//	iMaxHistory=0;
//	iWidth=0;
//	iHeight=0;
//	iPos=0;
//	iOrigin=0;
//	iLine=0;
//	iRecall=0;
//	iBuf=NULL;
//	iPrompt=_L("");
    __DECLARE_NAME(_S("CLineEdit"));
}


CLineEdit::~CLineEdit()
//
// Destroy the line editor
//
{

    TInt count=iHistory->Count();
    while (count--)
        User::Free((*iHistory)[count]);
    delete iHistory;
}

CLineEdit* CLineEdit::NewL(CAdvConsoleClient* aConsole,TInt aMaxHistory)
//
// Create a new line editor
//
{

    CLineEdit* pE=new(ELeave) CLineEdit;
    pE->iHistory=new(ELeave) CArrayFixFlat<HBufC*>(aMaxHistory+2);
    pE->iConsole=aConsole;
    pE->iMaxHistory=aMaxHistory;
    pE->iWidth=pE->iConsole->ScreenSize().iWidth;
    pE->iHeight=pE->iConsole->ScreenSize().iHeight;
    return(pE);
}

TInt CLineEdit::Lines()
//
// The number of lines being edited.
//
{

    TInt nL=1;
    TInt len = ApparentPos(Buf().Length());
    if (len >= iWidth-WIDTHKLUDGE-iOrigin)
        nL+=(len + iOrigin)/(iWidth-WIDTHKLUDGE);
    return(nL);
}

#define TABWIDTH 4
TInt CLineEdit::ApparentPos(int aPos)
// Given a buffer position 'aPos' work out the apparent position after
// expanding tabs.
{
    TInt pos = iOrigin;
    TInt line = 0;
    for (TInt i = 0; i < aPos; i++)
    {
        if (Buf()[i] == EKeyTab)
        {
            pos = (pos + TABWIDTH) / TABWIDTH * TABWIDTH;
        }
        else
            pos++;
        if (pos >= iWidth-WIDTHKLUDGE) pos = 0, line++;
    }
    pos += line * (iWidth-WIDTHKLUDGE);
    pos -= iOrigin;
    return pos;
}

TPoint CLineEdit::Where(int aPos)
//
// Return the real cursor position.
// This version tries to take account of any tabs.
//
{
    if (aPos == -1) aPos = iPos;
    TInt pos = ApparentPos(aPos);

    if (pos>=(iWidth-WIDTHKLUDGE-iOrigin))
        return(TPoint((pos+iOrigin)%(iWidth-WIDTHKLUDGE),((pos+iOrigin)/(iWidth-WIDTHKLUDGE))+iLine));
    return(TPoint(pos+iOrigin,iLine));
}

void CLineEdit::ClearLine()
//
// Clears the line being edited.
//
{
    if (Buf().Length())
    {
        TInt nL=Lines();
        while (nL--)
        {
            iConsole->SetPos(nL ? 0 : iOrigin,iLine+nL);
            iConsole->ClearToEndOfLine();
        }
        Buf().Zero();
        iPos=0;
    }
}

void CLineEdit::ClearLast(TInt aCnt)
//
// Clears the last aCnt characters.
//
{

    TInt aPos=iPos;
    iPos=((TInt)Buf().Length())-aCnt;
    while (iPos<((TInt)Buf().Length()))
    {
        TPoint p=Where();
        iConsole->SetCursorPosAbs(p);
        iConsole->ClearToEndOfLine();
        iPos+=(iWidth-p.iX);
    }
    iPos=aPos;
}

void CLineEdit::Recall()
//
// Recall a line for editing.
//
{
    if (iRecall!=(-1))
    {
        ClearLine();
        HBufC* pL=(*iHistory)[iRecall];
        Buf()=(*pL);
		Refresh();
        //iConsole->Write(Buf());
        iPos=Buf().Length();
        TInt nL=Lines();
        if ((iLine+nL)>(iHeight-HEIGHTKLUDGE))
            iLine=iHeight-HEIGHTKLUDGE-nL;
		Refresh();
        //Buf()=(*pL);
    }
}

TInt CLineEdit::WordLeft()
//
// Position the cursor to the next word left.
//
{
    TInt x=iPos-1;
    while (x && TChar(Buf()[x]).IsSpace())
        x--;
    while (x && TChar(Buf()[x]).IsGraph())
        x--;
    if (TChar(Buf()[x]).IsSpace())
        x++;
    return(x);
}

TInt CLineEdit::WordRight()
//
// Position the cursor to the next word right.
//
{
    TInt x=iPos;
    while (x<(TInt)Buf().Length() && TChar(Buf()[x]).IsGraph())
        x++;
    while (x<(TInt)Buf().Length() && TChar(Buf()[x]).IsSpace())
        x++;
    return(x);
}

void CLineEdit::Cursor()
//
// Position the cursor.
//
{
    iConsole->SetCursorPosAbs(Where());
}

void CLineEdit::Refresh()
//
// Refresh the line.
//
{
    iConsole->SetCursorHeight(ECursorNone);
    iConsole->SetPos(iOrigin,iLine);

    TInt left = 0, right = 0;
    while (right < Buf().Length())
    {
        while (right < Buf().Length() && Buf()[right] != '\t')
            right++;

        iConsole->Write(Buf().Mid(left, right - left));
        // Now expand the tab.
        if (right < Buf().Length())
        {
            right = left = right + 1;
            TInt xpos = iConsole->WhereX();
            TInt TabLen = (xpos + TABWIDTH) / TABWIDTH * TABWIDTH;
			if (TabLen > iWidth - WIDTHKLUDGE) TabLen = iWidth - WIDTHKLUDGE;
			TabLen -= xpos;
            TPtrC tab;
            tab.Set((const TUint8 *)"        ", TabLen);
            iConsole->Write(tab);
        }
    }
//    iConsole->Write(Buf());
    iConsole->ClearToEndOfLine();
    Cursor();
    iConsole->SetCursorHeight(iMode==EEditOverWrite ? ECursorNormal : ECursorInsert);
}

void CLineEdit::Edit(const TDesC& aPrompt,TDes* aBuf)
//
// Start the editor or a single key fetch.
//
{

    iBuf=aBuf;
    iPos=0;
    Buf().Zero();
    iMode=EEditInsert;
    iConsole->SetCursorHeight(iMode==EEditOverWrite ? ECursorNormal : ECursorInsert);
    iConsole->Write(aPrompt);
    iOrigin=iConsole->WhereX();
    iLine=iConsole->WhereY();
    iRecall=(-1);
    TInt hCount=iHistory->Count();
    if (hCount>iMaxHistory)
        hCount=iMaxHistory;
    FOREVER
    {
        TChar gChar=iConsole->Getch();
        switch (gChar)
        {
        case EKeyPointer:
            break;
        case EKeyEscape:
            ClearLine();
            iRecall=(-1);
            break;
        case EKeyHome:
            iPos=0;
            Cursor();
            break;
        case EKeyLeftArrow:
            if (iPos)
            {
                if((iConsole->KeyModifiers() & EModifierCtrl) != 0)
                    iPos=WordLeft();
                else
                    iPos--;
                Cursor();
            }
            break;
        case EKeyRightArrow:
            if (iPos<((TInt)Buf().Length()))
            {
                if((iConsole->KeyModifiers() & EModifierCtrl) != 0)
                    iPos=WordRight();
                else
                    iPos++;
                Cursor();
            }
            break;
        case EKeyEnd:
            iPos=((TInt)Buf().Length());
            Cursor();
            break;
        case EKeyPageUp:
            if (hCount==0)
                break;
            iRecall=hCount-1;
            Recall();
            break;
        case EKeyUpArrow:
            if (iRecall==(-1))
            {
                if (hCount==0)
                    break;
                iRecall=0;
            }
            else if (iRecall>=(hCount-1))
            {
                ClearLine();
                iRecall=(-1);
                break;
            }
            else
                iRecall++;
            Recall();
            break;
        case EKeyDownArrow:
            if (iRecall==(-1))
            {
                if (hCount==0)
                    break;
                iRecall=hCount-1;
            }
            else if (iRecall==0)
            {
                ClearLine();
                iRecall=(-1);
                break;
            }
            else
                iRecall--;
            Recall();
            break;
        case EKeyPageDown:
            if (hCount==0)
                break;
            iRecall=0;
            Recall();
            break;
        case EKeyEnter:
            iConsole->SetCursorHeight(ECursorNone);
            iConsole->SetPos(0,iLine+Lines()-1); // CR on the last line
            iConsole->Write(_L("\n")); // Just a line feed
//			iConsole->SetCursorHeight(iDefaultMode==EEditOverWrite ? ECursorNormal : ECursorInsert);
            iRecall=(-1);
            if (Buf().Length()>2)
            {
                if (iHistory->Count()==iMaxHistory+1)
                {
                    User::Free((*iHistory)[iMaxHistory]);
                    iHistory->Delete(iMaxHistory);
                }
                HBufC* pB=Buf().Alloc();
                if (pB!=NULL)
                    {TRAPD(r,iHistory->InsertL(0,pB));}
            }
            return;
        case CTRL('z'): case CTRL('d'): // EOF characters, only work at start of line.
            if (Buf().Length() == 0)
            {
                Buf().Append(gChar);
                return;
            }
            break;
        case EKeyBackspace:
            if ((iConsole->KeyModifiers() & EModifierShift) == 0)
            {
                if (iPos)
                {
                    TInt iN=1;
                    if ((iConsole->KeyModifiers() & EModifierCtrl) != 0)
                        iN=iPos-WordLeft();
                    ClearLast(iN);
                    iPos-=iN;
                    Buf().Delete(iPos,iN);
                    Refresh();
                }
                break;
            }
            // If backspace if shifted, fall through to delete key instead.

        case EKeyDelete:
            if (iPos<((TInt)Buf().Length()))
            {
                TInt iN=1;
                if ((iConsole->KeyModifiers() & EModifierCtrl) != 0)
                    iN=WordRight()-iPos;
                ClearLast(iN);
                Buf().Delete(iPos,iN);
                Refresh();
            }
            break;
        case EKeyInsert: case CTRL('v'):
            iMode=(iMode==EEditOverWrite ? EEditInsert : EEditOverWrite);
            iConsole->SetCursorHeight(iMode==EEditOverWrite ? ECursorNormal : ECursorInsert);
            break;
        default:
            if (!(gChar.IsPrint() || gChar == EKeyTab))
                break;
            if (iMode==EEditOverWrite && iPos<((TInt)Buf().Length()))
                Buf()[iPos++]=(TText)gChar;
            else if (Buf().Length()<KMaxCommandLine)
            {
                TInt oL=Lines();
                TBuf<0x02> b;
                b.Append(gChar);
                Buf().Insert(iPos++,b);
                TInt nL=Lines();
                if (nL!=oL)
                {
                    iConsole->SetCursorHeight(ECursorNone);
                    iConsole->SetPos(0,iLine+oL-1);
                    while (oL++ != nL)
                        iConsole->Write(_L("\n"));
                    iConsole->SetPos(0,iLine);
                    if (iLine + nL > iHeight-HEIGHTKLUDGE)
                        iLine = iHeight-HEIGHTKLUDGE - nL;
                }
            }
            else
            {
                iConsole->Write(_L("\7"));
                iConsole->SetPos((iOrigin+iPos)%(iWidth-WIDTHKLUDGE),iLine+Lines()-1);
                break;
            }
            Refresh();
        }
    }
}

class CTtyWindow : public CBase
{
public:
    CTtyWindow(const TDesC& aCaption);       // Constructor.
    void ConstructL(const TDesC *aFaceName=NULL);
	~CTtyWindow();
public:
    int read(char *buf, int n);
    int write(const char *buf, int n);
    fpos_t seek(fpos_t off, int whence);
    int close();

public:
    static int readfn(void *cookie, char *buf, int n);
    static int writefn(void *cookie, const char *buf, int n);
    static fpos_t seekfn(void *cookie, fpos_t off, int whence);
    static int closefn(void *cookie);

    CLineEdit *iCommandLine;
    TBool iInitialised;
    HBufC *iCaption;
    TInt refcount;
    CAdvConsoleClient *iConsole;
};

CTtyWindow::~CTtyWindow()
{
	if (iCaption) delete iCaption;
	if (iCommandLine) delete iCommandLine;
	if (iConsole) delete iConsole;
}

static CTtyWindow *sTty;

int CTtyWindow::readfn(void *cookie, char *buf, int n)
{
    return STATIC_CAST(CTtyWindow *, cookie)->read(buf, n);
}

int CTtyWindow::writefn(void *cookie, const char *buf, int n)
{
    return STATIC_CAST(CTtyWindow *, cookie)->write(buf, n);
}

fpos_t CTtyWindow::seekfn(void *cookie, fpos_t off, int whence)
{
    return STATIC_CAST(CTtyWindow *, cookie)->seek(off, whence);
}

int CTtyWindow::closefn(void *cookie)
{
    return STATIC_CAST(CTtyWindow *, cookie)->close();
}

CTtyWindow::CTtyWindow(const TDesC& aCaption): iInitialised(0), refcount(0)
{
    iCaption = HBufC::New(aCaption.Length());
    (*iCaption) = aCaption;
}

// Complete the construction of the console window.
// called when the first read or write occurs if not before.
void CTtyWindow::ConstructL(const TDesC *aFaceName)
{
    if (iInitialised) return;
    iInitialised = 1;
    iConsole = NewAdvConsole(aFaceName);
    ((CConsoleBase *)iConsole)->Create(*iCaption, TSize(80, 25));
    iCommandLine = CLineEdit::NewL(iConsole, 10);
}

int CTtyWindow::read(char *buf, int n)
{
    if (!iInitialised) ConstructL();
    // If we are reading at least 3 characters, make it line buffered.
    // otherwise just do a console read. A kludge I know, but it should work.
    if (n >= 3)
    {
        TPtr p((unsigned char*)buf, n-2);
        iCommandLine->Edit(_L(""), &p);
        if (p.Length() > 0 && (p[0] == CTRL('z') || p[0] == CTRL('d')))    // EOF markers, return 0 bytes.
        {
            p.SetLength(0);
        }
        else
        {
            p.Append('\n');
        }
        p.ZeroTerminate();
        return p.Length();
    }
    else
    {
        buf[0] = sTty->iConsole->Getch();
        return 1;
    }
}

int CTtyWindow::write(const char *buf, int n)
{
    if (!iInitialised) ConstructL();
    iConsole->Write(TPtrC((const TUint8 *)buf, n));
    return n;
}

fpos_t CTtyWindow::seek(fpos_t /*off*/, int /*whence*/)
{
    return 0;
}

int CTtyWindow::close()
{
    refcount--;
    if (refcount == 0)
    {
        iConsole->Write(_L("Press a key to exit"));
        iConsole->Getch();
        delete sTty;
        sTty = NULL;
    }
    return 0;
}

/* Replace the original stdin, stdout, stderr descriptors with my own version. */
void ReplaceStdio(const char *aCaption)
{
    CTtyWindow *newstdio = new (ELeave) CTtyWindow(TPtrC((const TUint8 *)aCaption));
    sTty = newstdio;
    //sTty->ConstructL();       // Delay construction in case we want
                                //to override the font.
    MagicFile(&ImpurePtr()->_sf[0], STATIC_CAST(void *, newstdio), CTtyWindow::readfn, (int (*)(void*,const char*,int))0, (fpos_t(*)(void*,fpos_t,int))0, CTtyWindow::closefn);
    MagicFile(&ImpurePtr()->_sf[1], STATIC_CAST(void *, newstdio), (int (*)(void*,char*,int))0, CTtyWindow::writefn, (fpos_t(*)(void*,fpos_t,int))0, CTtyWindow::closefn);
    MagicFile(&ImpurePtr()->_sf[2], STATIC_CAST(void *, newstdio), (int (*)(void*,char*,int))0, CTtyWindow::writefn, (fpos_t(*)(void*,fpos_t,int))0, CTtyWindow::closefn);
    newstdio->refcount = 3;
    ImpurePtr()->__sdidinit = 1;
    //sTty->write("Testing\n", 8);
#if 0
    for (int j = 0x0; j < 0x100; j += 0x10)
    {
        for (int i = 0; i < 16; i++)
        {
            static char *colors[] = {
                "black ", "d-grey ", "red ", "brown ", "green ",
                "b-red ", "yellow ", "b-green ", "b-magenta ", "b-blue ",
                "b-cyan ", "magenta ", "blue ", "cyan ", "grey ", "white "
            };
            fflush(stdout);
            sTty->iConsole->SetAttr(i | j | ATT_COLORMASK);
            xputs(colors[i]);
        }
    }
    xputs("\n");
    tgetch();
#endif
}

// Override the default font settings.
// This must be called before any output is generated as it relies
// on setting a global to be used when constructing the console.
extern "C" void EpocSetFont(const char *aName)
{
	TPtrC face;
	face.Set((const unsigned char *)aName, strlen(aName));
        if (!sTty->iInitialised) sTty->ConstructL(&face);
}

extern "C" {
    int tgetch();
    int tty_nh_poskey(int *x, int *y, int *mod);
    int mytty_kbhit();
    void get_scr_size();
    void nocmov(int x, int y);
    void cmov(int x, int y);
    void xputc(char c);
    void xputs(const char *s);
    void cl_end();
    void clear_screen();
    void home();
    void backsp();
    void tty_nhbell();
    void tty_delay_output();
    void cl_eos();
}

/*
 * return a key, or 0, in which case a mouse button was pressed
 * mouse events should be returned as character postitions in the map window.
 * Since normal tty's don't have mice, just return a key.
 */
/*ARGSUSED*/
int tty_nh_poskey(int *x, int *y, int *mod)
{
    int res = tty_nhgetch();
    while (res == 255)
    {
        *x = cursor_x + 1;      // Not sure why we need to adjust the x position.
        *y = cursor_y - 1;      // Want position on map window, so subtract 1 for top line.
        *mod = cursor_click;
        res = 0;
        if (*x < 0 || *y < 0 || *x >= COLNO || *y >= ROWNO)
            res = tty_nhgetch();
        else
            break;
    }
    return res;
}

int tgetch()
{
    // Read a key and translate according to the shift state.
    TChar gChar = sTty->iConsole->Getch();
    TInt shifts = sTty->iConsole->KeyModifiers();
    TBool Meta = (shifts & (EModifierFunc | EModifierAlt)) != 0;
    TBool Ctrl = (shifts & EModifierCtrl) != 0;
    TBool Shift = (shifts & EModifierShift) != 0;

    // Pointer clicks save the position in the map window, and pass back up
    // the call tree.
    if (gChar == EKeyPointer)
    {
        TPoint pointer = sTty->iConsole->PointerPos();
        cursor_x = pointer.iX;
        cursor_y = pointer.iY;
        cursor_click = Shift?CLICK_2:CLICK_1;
        return 255;
    }

    // shiftstate iConsole->KeyModifiers
    if (iflags.num_pad)
    {
        switch (gChar)
        {
        default:
            if (Meta && isalnum(gChar)) return M(gChar);
            return gChar;
        case EKeyEnter:     return C('j');
        case EKeyHome:      gChar = '7'; break;
        case EKeyLeftArrow: gChar = '4'; break;
        case EKeyRightArrow: gChar = '6'; break;
        case EKeyEnd:       gChar = '3'; break;
        case EKeyPageUp:    gChar = '9'; break;
        case EKeyPageDown:  gChar = '1'; break;
        case EKeyUpArrow:   gChar = '8'; break;
        case EKeyDownArrow: gChar = '2'; break;
        case EKeyMenu:      return C('z');
        }
        if (Ctrl) return M(gChar);
    }
    else
    {
        switch (gChar)
        {
        default:
            if (Meta && isalnum(gChar)) return M(gChar);
            return gChar;
        case EKeyEnter:     return C('j');
        case EKeyHome:      gChar = 'y'; break;
        case EKeyLeftArrow: gChar = 'h'; break;
        case EKeyRightArrow: gChar = 'l'; break;
        case EKeyEnd:       gChar = 'n'; break;
        case EKeyPageUp:    gChar = 'u'; break;
        case EKeyPageDown:  gChar = 'b'; break;
        case EKeyUpArrow:   gChar = 'k'; break;
        case EKeyDownArrow: gChar = 'j'; break;
        case EKeyMenu:      return C('z');
        }
        if (Shift) return tolower(gChar);
        if (Ctrl) return C(gChar);
    }
    return gChar;
}

int mytty_kbhit()
{
    // Has a key been hit?
     return sTty->iConsole->KeyHit() != 0;
}


/*
 * Called after returning from ! or ^Z
 */
void gettty()
{
	iflags.cbreak = TRUE;
}


void get_scr_size()
{
    if (!sTty->iInitialised) sTty->ConstructL();
    CO = sTty->iConsole->ScreenSize().iWidth - WIDTHKLUDGE;
    LI = sTty->iConsole->ScreenSize().iHeight - HEIGHTKLUDGE;
}

void nocmov(int x, int y)
{
    fflush(stdout);
    sTty->iConsole->SetPos(x, y);
}

void cmov(int x, int y)
{
    fflush(stdout);
    sTty->iConsole->SetPos(x, y);
    ttyDisplay->cury = y;
    ttyDisplay->curx = x;
}

void xputc(char c)
{
    TPtrC p((TUint8*)&c, 1);
    fflush(stdout);
    sTty->iConsole->Write(p);
}

void xputs(const char *s)
{
    TPtrC p((const TUint8*)s, strlen(s));
    fflush(stdout);
    if (!sTty->iInitialised) sTty->ConstructL();
    sTty->iConsole->Write(p);
}

void cl_end()
{
    fflush(stdout);
    sTty->iConsole->ClearToEndOfLine();
    tty_curs(BASE_WINDOW, (int)ttyDisplay->curx+1, (int)ttyDisplay->cury);
}


void clear_screen()
{
    fflush(stdout);
    sTty->iConsole->ClearScreen();
    home();
}


void home()
{
    fflush(stdout);
    tty_curs(BASE_WINDOW, 1, 0);
    ttyDisplay->curx = ttyDisplay->cury = 0;
}

void
   term_end_raw_bold(void)
{
    fflush(stdout);
    sTty->iConsole->SetAttr(ATT_NORMAL);
}

void
   term_start_raw_bold(void)
{
    fflush(stdout);
    sTty->iConsole->SetAttr(ATT_BOLD);
}

void standoutbeg()
{
    term_start_attr(ATR_INVERSE);
}

void standoutend()
{
    term_end_attr(ATR_INVERSE);
}

void term_end_attr(int attr)
{
    if (last_colour == NO_COLOR)
    {
        fflush(stdout);
        sTty->iConsole->SetAttr(ATT_NORMAL);
    }
    else
        term_start_color(last_colour);
}

void term_start_attr(int attr)
{
    fflush(stdout);
    switch(attr){
    case ATR_ULINE:
        sTty->iConsole->SetAttr(ATT_UNDERLINE);
        break;
    case ATR_BOLD:
        sTty->iConsole->SetAttr(ATT_BOLD);
        break;
    case ATR_BLINK:
    case ATR_INVERSE:
        if (last_colour==NO_COLOR)
        {
            sTty->iConsole->SetAttr(ATT_INVERSE);
        }
        else
        {
            // Inverse video when we have a colour selected simply tries to
            // invert the background colour.
            int color = (EpocColorDisplay?InverseColorTable:InverseMonoTable)[last_colour];
            sTty->iConsole->SetAttr(color);
        }
        break;
    default:
        term_end_attr(0);
        break;
	}                
}


int
   has_color(int color)
{
	return 1;
}

void
   term_start_color(int color)
{
# ifdef TEXTCOLOR

    if (color < 0 || color >= sizeof(ColorTable))
        color = sizeof(ColorTable)-1;

    last_colour = color;
    color = (EpocColorDisplay?ColorTable:MonoTable)[color];
    fflush(stdout);
    sTty->iConsole->SetAttr(color);
# endif
}

void
   term_end_color(void)
{
    fflush(stdout);
    sTty->iConsole->SetAttr(ATT_NORMAL);
    last_colour = NO_COLOR;
}



void backsp()
{
    fflush(stdout);
    int x = sTty->iConsole->WhereX();
    if (x > 0) x-= 1;
    sTty->iConsole->SetPos(x);
    sTty->iConsole->Write(_L(" "));
    sTty->iConsole->SetPos(x);
}

void tty_nhbell()
{
    const TInt KFrequency=8000; // Hertz
    const TInt KDuration=500000;        // Microseconds

    if (flags.silent) return;
    // Beep(8000,500);
    User::Beep(KFrequency,TTimeIntervalMicroSeconds32(KDuration));
}

void tty_delay_output()
{
    User::After(50000); // Sleep for 50mS.
}

void cl_eos()
{
    fflush(stdout);
    register int cy = ttyDisplay->cury+1;
    while(cy <= LI-2) {
        cl_end();
        xputc('\n');
        cy++;
    }
    cl_end();
    tty_curs(BASE_WINDOW, (int)ttyDisplay->curx+1, (int)ttyDisplay->cury);
}
